// kate: replace-tabs off; indent-width 4; indent-mode normal
// vim: ts=4:sw=4:noexpandtab
/*

Copyright (c) 2010--2012,
François Pomerleau and Stephane Magnenat, ASL, ETHZ, Switzerland
You can contact the authors at <f dot pomerleau at gmail dot com> and
<stephane at magnenat dot net>

All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the <organization> nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL ETH-ASL BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/

#include "Bibliography.h"

#include <boost/lexical_cast.hpp>
#include <iostream>
#include <cassert>

namespace PointMatcherSupport
{
	using namespace std;
	
	static Bibliography bibliography()
	{
		return Bibliography({
			{ "Phillips2007VarTrimmed",
				StringMap({
					{ "type", "inproceedings" },
					{ "title", "Outlier robust ICP for minimizing fractional RMSD" },
					{ "author", "Phillips, J.M. and Liu, R. and Tomasi, C." },
					{ "booktitle", "3-D Digital Imaging and Modeling, 2007. 3DIM '07. Sixth International Conference on" },
					{ "year", "2007" },
					{ "pages", "427--434" },
					{ "publisher", "IEEE Press" },
					{ "doi", "10.1109/3DIM.2007.39" },
					{ "fulltext", "http://x86.cs.duke.edu/~tomasi/papers/phillips/phillips3DIM07.pdf" }
				})
			},
			{ "Chetverikov2002Trimmed", 
				StringMap({
					{ "type", "inproceedings" },
					{ "title", "The Trimmed Iterative Closest Point Algorithm" },
					{ "author", "Chetverikov, D. and Svirko, D. and Stepanov, D. and Krsek, P." },
					{ "booktitle", "Pattern Recognition, 2002. Proceedings. 16th International Conference on" },
					{ "year", "2002" },
					{ "pages", "545--548" },
					{ "publisher", "IEEE Press" },
					{ "doi", "10.1109/ICPR.2002.1047997 " },
					{ "fulltext", "http://hci.iwr.uni-heidelberg.de/publications/dip/2002/ICPR2002/DATA/10_1_03.PDF"}
				})
			},
			{ "Besl1992Point2Point", 
				StringMap({
					{ "type", "inproceedings" },
					{ "title", "A Method for Registration of 3-D Shapes" },
					{ "author", "Besl, P.J. and McKay, H.D." },
					{ "booktitle", "Pattern Analysis and Machine Intelligence, IEEE Transactions" },
					{ "year", "1992" },
					{ "pages", "239--256" },
					{ "publisher", "IEEE Press" },
					{ "doi", "10.1109/34.121791" },
					{ "fulltext", "http://ieeexplore.ieee.org/xpls/abs_all.jsp?arnumber=121791"}
				})
			},
			{ "Chen1991Point2Plane", 
				StringMap({
					{ "type", "inproceedings" },
					{ "title", "Object modeling by registration of multiple range images" },
					{ "author", "Chen, Y. and Medioni, G." },
					{ "booktitle", "Robotics and Automation, 1991. Proceedings., 1991 IEEE International Conference on" },
					{ "year", "1991" },
					{ "pages", "2724--2729" },
					{ "publisher", "IEEE Press" },
					{ "doi", "10.1109/ROBOT.1991.132043" },
					{ "fulltext", "http://ieeexplore.ieee.org/search/srchabstract.jsp?tp=&arnumber=132043"}
				})
			},
			{ "Masuda1996Random", 
				StringMap({
					{ "type", "inproceedings" },
					{ "title", "Registration and integration of multiple range images for 3-D model construction" },
					{ "author", "Masuda, T. and Sakaue, K. and Yokoya, N." },
					{ "booktitle", "Pattern Recognition, 1996., Proceedings of the 13th International Conference on" },
					{ "year", "1996" },
					{ "pages", "879--883" },
					{ "publisher", "IEEE Press" },
					{ "doi", "10.1109/ICPR.1996.546150" },
					{ "fulltext", "http://ieeexplore.ieee.org/xpls/abs_all.jsp?arnumber=546150"}
				})
			},
			{ "Diebel2004Median", 
				StringMap({
					{ "type", "inproceedings" },
					{ "title", "Simultaneous Localization and Mapping with Active Stereo Vision" },
					{ "author", "Diebel, J. and Reutersward, K. and Thrun, S. and Davis, J. and Gupta, R." },
					{ "booktitle", "Intelligent Robots and Systems, 2004. (IROS 2004). Proceedings. 2004 IEEE/RSJ International Conference on" },
					{ "year", "2004" },
					{ "pages", "3436--3443" },
					{ "publisher", "IEEE Press" },
					{ "doi", "10.1109/IROS.2004.1389948" },
					{ "fulltext", "http://ieeexplore.ieee.org/xpls/abs_all.jsp?arnumber=1389948"}
				})
			}

		});
	}
	
	CurrentBibliography::CurrentBibliography(Mode mode):
		mode(mode)
	{}
	
	void CurrentBibliography::dump(std::ostream& os) const
	{
		switch (mode)
		{
			case NORMAL: dumpText(os); break;
			case ROSWIKI: dumpWiki(os); break;
			case BIBTEX: dumpBibtex(os); break;
			default: assert(false); break;
		};
	}
	
	void CurrentBibliography::dumpText(std::ostream& os) const
	{
		Bibliography biblio(bibliography());
		for (size_t i = 0; i < entries.size(); ++i)
		{
			const string& entryName(entries[i]);
			if (!biblio.contains(entryName))
				throw runtime_error(string("Broken bibliography, missing entry " + entryName));
			const StringMap& entry(biblio.get(entryName));
			
			os << "[" << i+1 << "]";
			if (entry.contains("title"))
				os << " " << entry.get("title") << ".";
			if (entry.contains("author"))
				os << " " << entry.get("author") << "";
			if (entry.contains("booktitle"))
				os << " In " << entry.get("booktitle") << ".";
			if (entry.contains("journal"))
				os << " " << entry.get("journal") << ".";
			if (entry.contains("pages"))
				os << " " << entry.get("pages") << ".";
			if (entry.contains("year"))
				os << " " << entry.get("year") << ".";
			os << endl << endl;
		}
	}

	void CurrentBibliography::dumpWiki(std::ostream& os) const
	{
		Bibliography biblio(bibliography());
		for (size_t i = 0; i < entries.size(); ++i)
		{
			const string& entryName(entries[i]);
			if (!biblio.contains(entryName))
				throw runtime_error(string("Broken bibliography, missing entry " + entryName));
			const StringMap& entry(biblio.get(entryName));
			
			os << " * " << "<<Anchor(" << entryName << ")>>[" << i+1 << "] -";
			if (entry.contains("title"))
				os << " '''" << entry.get("title") << ".'''";
			if (entry.contains("author"))
				os << " " << entry.get("author") << "";
			if (entry.contains("booktitle"))
				os << " ''In " << entry.get("booktitle") << ".''";
			if (entry.contains("journal"))
				os << " " << entry.get("journal") << ".";
			if (entry.contains("pages"))
				os << " " << entry.get("pages") << ".";
			if (entry.contains("year"))
				os << " " << entry.get("year") << ".";
			if (entry.contains("doi"))
				os << " DOI: [[http://dx.doi.org/" << entry.get("doi") << "|" << entry.get("doi") << "]].";
			if (entry.contains("fulltext"))
				os << " [[" << entry.get("fulltext") << "|full text]].";
			os << endl;
		}
	}

	void CurrentBibliography::dumpBibtex(std::ostream& os) const
	{
		Bibliography biblio(bibliography());
		for (size_t i = 0; i < entries.size(); ++i)
		{
			const string& entryName(entries[i]);
			if (!biblio.contains(entryName))
				throw runtime_error(string("Broken bibliography, missing entry " + entryName));
			const StringMap& entry(biblio.get(entryName));
			
			os << "@" << entry.get("type") << "{" << entryName << endl;
			if (entry.contains("title"))
				os << "\ttitle={" << entry.get("title") << "}," << endl;
			if (entry.contains("author"))
				os << "\tauthor={" << entry.get("author") << "}," << endl;
			if (entry.contains("booktitle"))
				os << "\tbooktitle={" << entry.get("booktitle") << "}," << endl;
			if (entry.contains("journal"))
				os << "\tjournal={" << entry.get("journal") << "}," << endl;
			if (entry.contains("pages"))
				os << "\tpages={" << entry.get("pages") << "}," << endl;
			if (entry.contains("year"))
				os << "\tyear={" << entry.get("year") << "}," << endl;
			os << "}" << endl << endl;
		}
	}
	
	static StringVector splitString(const string& text, char delim)
	{
		StringVector res;
		size_t pos = 0;
		while(true)
		{
			const size_t nextPos = text.find(delim, pos);
			if (nextPos == text.npos)
			{
				res.push_back(text.substr(pos));
				break;
			}
			else
				res.push_back(text.substr(pos, nextPos - pos));
			pos = nextPos + 1;
		}
		return res;
	}
	
	std::string getAndReplaceBibEntries(const std::string& text, CurrentBibliography& curBib)
	{
		CurrentBibliography::Mode mode(curBib.mode);
		BibIndices& indices(curBib.indices);
		StringVector& entries(curBib.entries);
		
		string newText;
		const StringVector words(splitString(text, ' '));
		for (size_t i = 0; i < words.size(); ++i)
		{
			const string& word(words[i]);
			const size_t l(word.length());
			const size_t p(word.find('}'));
			if ((l > 7) && (word.substr(0, 6) == "\\cite{") && (p != string::npos))
			{
				if (mode == CurrentBibliography::ROSWIKI)
					newText += "&#91;";
				else
					newText += '[';
				const StringVector keys(splitString(word.substr(6, p-6), ','));
				for (size_t j = 0; j < keys.size(); ++j)
				{
					const string key(keys[j]);
					if (mode == CurrentBibliography::ROSWIKI)
						newText += "[[#" + key + "|";
					if (indices.contains(key))
					{
						newText += boost::lexical_cast<string>(indices.get(key)+1);
					}
					else
					{
						size_t index(entries.size());
						entries.push_back(key);
						indices[key] = index;
						newText += boost::lexical_cast<string>(index+1);
					}
					if (mode == CurrentBibliography::ROSWIKI)
						newText += "]]";
					if (j+1 != keys.size())
						newText += ',';
				}
				if (mode == CurrentBibliography::ROSWIKI)
					newText += "&#93;";
				else
					newText += ']';
				newText += word.substr(p+1);
			}
			else
				newText += word;
			if (i+1 != words.size())
				newText += ' ';
		}
		if (mode == CurrentBibliography::BIBTEX)
			return text;
		else
			return newText;
	}
}; // PointMatcherSupport
